# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data models for db collections."""

import datetime as dt
import enum
import re
from collections.abc import Generator
from functools import cached_property
from typing import (
    Any,
    ClassVar,
    Optional,
    Self,
    TYPE_CHECKING,
    TypeAlias,
    cast,
    override,
)

import jsonpath_rw
from django.conf import settings
from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.fields import RangeOperators
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import models, transaction
from django.db.models import (
    CheckConstraint,
    Count,
    Deferrable,
    Exists,
    F,
    JSONField,
    Max,
    OuterRef,
    Q,
    QuerySet,
    UniqueConstraint,
    Value,
)
from django.db.models.fields.json import KT
from django.db.models.functions import Coalesce
from django.urls import reverse
from django.utils import timezone

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    BaseArtifactDataModel,
    CollectionCategory,
    DebusinePromise,
    SINGLETON_COLLECTION_CATEGORIES,
)
from debusine.db.constraints import JsonDataUniqueConstraint
from debusine.db.models.permissions import (
    Allow,
    Role,
    RoleBase,
    Roles,
    permission_check,
    permission_filter,
)
from debusine.db.models.scopes import Scope
from debusine.db.models.workspaces import Workspace
from debusine.db.permissioncontext import PermissionContext

if TYPE_CHECKING:
    from django.http import HttpRequest
    from django_stubs_ext.db.models import TypedModelMeta

    from debusine.db.models.artifacts import Artifact
    from debusine.db.models.auth import Token, User
    from debusine.db.models.work_requests import WorkRequest
    from debusine.server.collections import CollectionManagerInterface
    from debusine.web.views.ui.collection_relations import CollectionRelationUI
    from debusine.web.views.ui.collections import CollectionItemUI, CollectionUI
else:
    TypedModelMeta = object

#: Regexp matching the structure of collection names
collection_name_regex = re.compile(r"^[A-Za-z][A-Za-z0-9+._-]*$")


def is_valid_collection_name(value: str) -> bool:
    """Check if value is a valid scope name."""
    return value == "_" or bool(collection_name_regex.match(value))


def validate_collection_name(value: str) -> None:
    """Validate collection names."""
    if not is_valid_collection_name(value):
        raise ValidationError(
            "%(value)r is not a valid collection name", params={"value": value}
        )


class _CollectionRetainsArtifacts(models.TextChoices):
    """Choices for Collection.retains_artifacts."""

    NEVER = "never", "Never"
    WORKFLOW = "workflow", "While workflow is running"
    ALWAYS = "always", "Always"


class CollectionRoleBase(RoleBase):
    """Collection role implementation."""

    implied_by_scope_roles: frozenset[Scope.Roles]
    implied_by_workspace_roles: frozenset[Workspace.Roles]
    implied_by_collection_roles: frozenset["CollectionRoles"]
    implied_by_public: bool

    @override
    def _setup(self) -> None:
        """Set up implications for a newly constructed role."""
        implied_by_scope_roles: set[Scope.Roles] = set()
        implied_by_workspace_roles: set[Workspace.Roles] = set()
        implied_by_collection_roles: set[CollectionRoles] = {
            cast(CollectionRoles, self)
        }
        implied_by_public: bool = False
        for i in self.implied_by:
            match i:
                case Workspace.Roles():
                    implied_by_scope_roles |= i.implied_by_scope_roles
                    implied_by_workspace_roles |= i.implied_by_workspace_roles
                    implied_by_public = implied_by_public or i.implied_by_public
                case Role():
                    # Resolve a role passed during class definition into its
                    # enum instance
                    role = self.__class__(i.value)
                    implied_by_scope_roles |= role.implied_by_scope_roles
                    implied_by_workspace_roles |= (
                        role.implied_by_workspace_roles
                    )
                    implied_by_collection_roles |= (
                        role.implied_by_collection_roles
                    )
                    implied_by_public = (
                        implied_by_public or role.implied_by_public
                    )
                case _:
                    raise ImproperlyConfigured(
                        f"Collection roles do not support implications by {i!r}"
                    )
        self.implied_by_scope_roles = frozenset(implied_by_scope_roles)
        self.implied_by_workspace_roles = frozenset(implied_by_workspace_roles)
        self.implied_by_collection_roles = frozenset(
            implied_by_collection_roles
        )
        self.implied_by_public = implied_by_public

    def q(self, pc: PermissionContext) -> Q:
        """Return a Q expression to select collections with this role."""
        if not pc.user.is_authenticated:
            if self.implied_by_public:
                return Q(workspace__public=True)
            else:
                return Q(pk__in=())

        q = Q(
            roles__group__in=pc.groups(),
            roles__role__in=self.implied_by_collection_roles,
        )
        workspace_q = Q(
            roles__group__in=pc.groups(),
            roles__role__in=self.implied_by_workspace_roles,
        ) | Q(
            scope__in=Scope.objects.filter(
                roles__group__in=pc.groups(),
                roles__role__in=self.implied_by_scope_roles,
            )
        )
        if self.implied_by_public:
            workspace_q |= Q(public=True)
        q |= Q(workspace__in=Workspace.objects.filter(workspace_q))

        return q

    def implies(self, role: "CollectionRoles") -> bool:
        """Check if this role implies the given one."""
        return (
            self.implied_by_scope_roles <= role.implied_by_scope_roles
            and self.implied_by_workspace_roles
            <= role.implied_by_workspace_roles
            and self.implied_by_collection_roles
            <= role.implied_by_collection_roles
            and self.implied_by_public <= role.implied_by_public
        )


class CollectionRoles(Roles, CollectionRoleBase, enum.ReprEnum):
    """Available roles for a Collection."""

    OWNER = Role("owner", implied_by=[Workspace.Roles.OWNER])

    VIEWER = Role("viewer", implied_by=[OWNER, Workspace.Roles.VIEWER])


CollectionRoles.setup()


class CollectionQuerySet[A](QuerySet["Collection", A]):
    """Custom QuerySet for Collection."""

    def in_current_scope(self) -> Self:
        """Filter to collections in the current scope."""
        from debusine.db.context import context

        return self.filter(workspace__scope=context.require_scope())

    def in_current_workspace(self) -> Self:
        """Filter to collections in the current workspace."""
        from debusine.db.context import context

        return self.filter(workspace=context.require_workspace())

    def with_role(self, pc: PermissionContext, role: CollectionRoles) -> Self:
        """Keep only resources where the user has the given role."""
        return self.filter(
            Exists(self.model.objects.filter(role.q(pc), pk=OuterRef("pk")))
        )

    @permission_filter(
        work_request=Allow.PASS,
        anonymous=Allow.PASS,
        explicit_grants_role=CollectionRoles.VIEWER,
    )
    def can_display(self, pc: PermissionContext) -> Self:
        """Keep only Collections that can be displayed."""
        return self.with_role(pc, CollectionRoles.VIEWER)

    @permission_filter()
    def can_configure(self, pc: PermissionContext) -> Self:
        """Keep only Collections that can be configured."""
        return self.with_role(pc, CollectionRoles.OWNER)

    @permission_filter()
    def can_create_collection_relation(self, pc: PermissionContext) -> Self:
        """Keep only Collections that can have a collection relation added."""
        return self.with_role(pc, CollectionRoles.OWNER)

    @permission_filter()
    def can_delete(self, pc: PermissionContext) -> Self:
        """Keep only Collections that can be deleted."""
        return self.with_role(pc, CollectionRoles.OWNER).exclude(
            category=CollectionCategory.WORKFLOW_INTERNAL
        )

    def exported_suites(self) -> "CollectionQuerySet[Any]":
        """Filter to exported ``debian:suite`` collections."""
        return (
            self.filter(category=CollectionCategory.SUITE)
            .annotate(
                exported=Coalesce(
                    "data__exported", Value("true"), output_field=JSONField()
                )
            )
            .exclude(exported=False)
        )


class CollectionManager(models.Manager["Collection"]):
    """Manager for Collection model."""

    def get_roles_model(self) -> type["CollectionRole"]:
        """Get the model used for role assignment."""
        return CollectionRole

    @override
    def get_queryset(self) -> CollectionQuerySet[Any]:
        """Use the custom QuerySet."""
        return CollectionQuerySet(self.model, using=self._db)

    @staticmethod
    def get_or_create_singleton(
        category: CollectionCategory,
        workspace: Workspace,
        *,
        data: dict[str, Any] | None = None,
    ) -> "tuple[Collection, bool]":
        """Create a singleton collection."""
        if category not in SINGLETON_COLLECTION_CATEGORIES:
            raise ValueError(
                f"'{category}' is not a singleton collection category"
            )
        return Collection.objects.get_or_create(
            name="_", category=category, workspace=workspace, data=data or {}
        )


class Collection(models.Model):
    """Model representing a collection."""

    Roles: TypeAlias = CollectionRoles
    RetainsArtifacts: TypeAlias = _CollectionRetainsArtifacts

    name = models.CharField(
        max_length=255, validators=[validate_collection_name]
    )
    category = models.CharField(max_length=255)
    full_history_retention_period = models.DurationField(
        null=True,
        blank=True,
        help_text="Period after removing an item from the collection"
        " during which the related artifact is not allowed to expire",
    )
    metadata_only_retention_period = models.DurationField(
        null=True,
        blank=True,
        help_text="Period after removing an item from the collection during"
        " which its metadata is kept in the history of the collection",
    )

    workspace = models.ForeignKey(
        Workspace, on_delete=models.PROTECT, related_name="collections"
    )
    retains_artifacts = models.CharField(
        max_length=8,
        choices=RetainsArtifacts.choices,
        default=RetainsArtifacts.ALWAYS,
        help_text="Controls if artifacts in this collection should be retained"
        " rather than being allowed to expire",
    )

    child_artifacts = models.ManyToManyField(
        "db.Artifact",
        through="db.CollectionItem",
        through_fields=("parent_collection", "artifact"),
        related_name="parent_collections",
    )
    child_collections = models.ManyToManyField(
        "self",
        through="db.CollectionItem",
        through_fields=("parent_collection", "collection"),
        related_name="parent_collections",
        symmetrical=False,
    )

    data = models.JSONField(default=dict, blank=True)

    objects = CollectionManager.from_queryset(CollectionQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"
        constraints = [
            UniqueConstraint(
                fields=["name", "category", "workspace"],
                name="%(app_label)s_%(class)s_unique_name_category_workspace",
            ),
            CheckConstraint(
                check=~Q(name=""), name="%(app_label)s_%(class)s_name_not_empty"
            ),
            CheckConstraint(
                check=(
                    (
                        ~Q(category__in=sorted(SINGLETON_COLLECTION_CATEGORIES))
                        & ~Q(name__startswith="_")
                    )
                    | Q(
                        category__in=sorted(SINGLETON_COLLECTION_CATEGORIES),
                        name="_",
                    )
                ),
                name="%(app_label)s_%(class)s_name_not_reserved",
            ),
            CheckConstraint(
                check=~Q(category=""),
                name="%(app_label)s_%(class)s_category_not_empty",
            ),
        ]

    @override
    def __str__(self) -> str:
        """Return id, name, category."""
        # Stringify using a valid lookup syntax
        return f"{self.name}@{self.category}"

    @override
    def __repr__(self) -> str:
        """Return id, name, category."""
        # Stringify using a valid lookup syntax
        return f"<Collection: {self.name}@{self.category} ({self.pk})>"

    def get_absolute_url(self) -> str:
        """Return the canonical URL to display the collection."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.workspace.scope.name):
            return reverse(
                "workspaces:collections:detail",
                kwargs={
                    "wname": self.workspace.name,
                    "ccat": self.category,
                    "cname": self.name,
                },
            )

    def get_absolute_url_search(self) -> str:
        """Return the canonical URL to search inside the collection."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.workspace.scope.name):
            return reverse(
                "workspaces:collections:search",
                kwargs={
                    "wname": self.workspace.name,
                    "ccat": self.category,
                    "cname": self.name,
                },
            )

    @cached_property
    def manager(self) -> "CollectionManagerInterface":
        """Get collection manager for this collection category."""
        # Local import to avoid circular dependency
        from debusine.server.collections import CollectionManagerInterface

        return CollectionManagerInterface.get_manager_for(self)

    def ui(self, request: "HttpRequest") -> "CollectionUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.collections import CollectionUI

        return CollectionUI.get(request, self)

    def grant_role(
        self, token: "Token", role: CollectionRoles, *, created_by: "User"
    ) -> "CollectionTokenGrant":
        """Grant a role on this collection to a token."""
        assert token.explicit_grants_only
        grant, _ = CollectionTokenGrant.objects.get_or_create(
            token=token, collection=self, role=role, created_by=created_by
        )
        return grant

    def has_role(self, pc: PermissionContext, role: CollectionRoles) -> bool:
        """Check if the user has the given role on this Collection."""
        from debusine.db.context import context

        # Honor implied_by_public immediately as it's independent from the user
        if role.implied_by_public and self.workspace.public:
            return True

        if not pc.user.is_authenticated:
            return False

        if context.pc_is(pc) and context.workspace == self.workspace:
            # Allowed as implied by scope roles the user has
            if not context.scope_roles.isdisjoint(role.implied_by_scope_roles):
                return True

            # Allowed as implied by workspace roles the user has
            if not context.workspace_roles.isdisjoint(
                role.implied_by_workspace_roles
            ):
                return True

        return (
            Collection.objects.with_role(pc, role).filter(pk=self.pk).exists()
        )

    @permission_check(
        "{user} cannot display collection {resource}",
        work_request=Allow.PASS,
        anonymous=Allow.PASS,
        explicit_grants_role=CollectionRoles.VIEWER,
    )
    def can_display(self, pc: PermissionContext) -> bool:
        """Check if the collection can be displayed."""
        return self.has_role(pc, CollectionRoles.VIEWER)

    @permission_check(
        "{user} cannot configure collection {resource}",
    )
    def can_configure(self, pc: PermissionContext) -> bool:
        """Check if the collection can be configured."""
        return self.has_role(pc, CollectionRoles.OWNER)

    @permission_check(
        "{user} cannot delete collection {resource}",
    )
    def can_delete(self, pc: PermissionContext) -> bool:
        """Check if the collection can be deleted."""
        return (
            self.category != CollectionCategory.WORKFLOW_INTERNAL
            and self.has_role(pc, CollectionRoles.OWNER)
        )

    @permission_check(
        "{user} cannot create collection relation in {resource}",
    )
    def can_create_collection_relation(self, pc: PermissionContext) -> bool:
        """Check if a relation can be created from this collection."""
        return self.has_role(pc, CollectionRoles.OWNER)

    def get_item_stats(
        self,
    ) -> Generator[dict[str, Any]]:
        """Return a dict of item counts for each item type."""
        for rec in (
            self.child_items.values(
                "child_type", "category", active=Q(removed_at__isnull=True)
            )
            .annotate(count=Count("id"))
            .order_by("child_type", "category", "active")
        ):
            child_type = CollectionItem.Types(rec["child_type"])
            yield {
                "type": child_type,
                "category": rec["category"],
                "active": rec["active"],
                "count": rec["count"],
            }

    def delete_collection(self) -> None:
        """
        Delete this collection, and all its CollectionItem entries.

        Does not delete the associated Collections or Artifacts.
        """
        self.child_items.all().delete()
        self.delete()


class CollectionRole(models.Model):
    """Role assignment for collection."""

    resource = models.ForeignKey(
        Collection,
        on_delete=models.CASCADE,
        related_name="roles",
    )

    group = models.ForeignKey(
        "Group",
        on_delete=models.CASCADE,
        related_name="collection_roles",
    )

    role = models.CharField(max_length=16, choices=CollectionRoles.choices)

    class Meta(TypedModelMeta):
        constraints = [
            UniqueConstraint(
                fields=["resource", "group", "role"],
                name="%(app_label)s_%(class)s_unique_resource_group_role",
            ),
        ]


class CollectionItemQuerySet(QuerySet["CollectionItem"]):
    """Custom QuerySet for CollectionItem."""

    def active(self) -> "CollectionItemQuerySet":
        """Filter to only active collection items."""
        return super().filter(removed_at__isnull=True)

    def active_at(self, at: dt.datetime) -> "CollectionItemQuerySet":
        """Filter to collection items that were active at a given time."""
        return self.filter(
            Q(created_at__lte=at),
            Q(removed_at__isnull=True) | Q(removed_at__gt=at),
        )

    def artifacts_in_collection(
        self, parent_collection: Collection, category: ArtifactCategory
    ) -> "CollectionItemQuerySet":
        """
        Filter to artifacts in a given parent collection.

        This is tuned to use the indexes on :py:class:`CollectionItem`.
        """
        return self.filter(
            parent_collection=parent_collection,
            # Technically redundant with parent_collection, but clues
            # PostgreSQL into using the correct index.
            parent_category=parent_collection.category,
            child_type=CollectionItem.Types.ARTIFACT,
            category=category,
        )


class CollectionItemManager(models.Manager["CollectionItem"]):
    """Manager for CollectionItem model."""

    @staticmethod
    def create_from_bare_data(
        category: BareDataCategory,
        *,
        parent_collection: Collection,
        name: str,
        data: BaseArtifactDataModel | dict[str, Any],
        created_at: dt.datetime | None = None,
        created_by_user: "User",
        created_by_workflow: Optional["WorkRequest"],
        replaced_by: Optional["CollectionItem"] = None,
    ) -> "CollectionItem":
        """Create a CollectionItem from bare data."""
        if isinstance(data, BaseArtifactDataModel):
            data = data.model_dump(mode="json", exclude_unset=True)

        match category:
            case BareDataCategory.PROMISE:
                # Raise ValueError if data is not valid
                DebusinePromise(**data)

        kwargs: dict[str, Any] = {}
        if created_at is not None:
            kwargs["created_at"] = created_at
        if replaced_by is not None:
            kwargs["removed_at"] = replaced_by.created_at
            kwargs["removed_by_user"] = replaced_by.created_by_user
            kwargs["removed_by_workflow"] = replaced_by.created_by_workflow

        return CollectionItem.objects.create(
            parent_collection=parent_collection,
            name=name,
            child_type=CollectionItem.Types.BARE,
            category=category,
            data=data,
            created_by_user=created_by_user,
            created_by_workflow=created_by_workflow,
            **kwargs,
        )

    @staticmethod
    def create_from_artifact(
        artifact: "Artifact",
        *,
        parent_collection: Collection,
        name: str,
        data: dict[str, Any],
        created_at: dt.datetime | None = None,
        created_by_user: "User",
        created_by_workflow: Optional["WorkRequest"] = None,
        replaced_by: Optional["CollectionItem"] = None,
    ) -> "CollectionItem":
        """Create a CollectionItem from the artifact."""
        kwargs: dict[str, Any] = {}
        if created_at is not None:
            kwargs["created_at"] = created_at
        if replaced_by is not None:
            kwargs["removed_at"] = replaced_by.created_at
            kwargs["removed_by_user"] = replaced_by.created_by_user
            kwargs["removed_by_workflow"] = replaced_by.created_by_workflow

        return CollectionItem.objects.create(
            parent_collection=parent_collection,
            name=name,
            artifact=artifact,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact.category,
            data=data,
            created_by_user=created_by_user,
            created_by_workflow=created_by_workflow,
            **kwargs,
        )

    @staticmethod
    def create_from_collection(
        collection: Collection,
        *,
        parent_collection: Collection,
        name: str,
        data: dict[str, Any],
        created_at: dt.datetime | None = None,
        created_by_user: "User",
        created_by_workflow: Optional["WorkRequest"] = None,
        replaced_by: Optional["CollectionItem"] = None,
    ) -> "CollectionItem":
        """Create a CollectionItem from the collection."""
        kwargs: dict[str, Any] = {}
        if created_at is not None:
            kwargs["created_at"] = created_at
        if replaced_by is not None:
            kwargs["removed_at"] = replaced_by.created_at
            kwargs["removed_by_user"] = replaced_by.created_by_user
            kwargs["removed_by_workflow"] = replaced_by.created_by_workflow

        return CollectionItem.objects.create(
            parent_collection=parent_collection,
            name=name,
            collection=collection,
            child_type=CollectionItem.Types.COLLECTION,
            category=collection.category,
            data=data,
            created_by_user=created_by_user,
            created_by_workflow=created_by_workflow,
            **kwargs,
        )

    def drop_full_history(self, at: dt.datetime) -> None:
        """
        Drop artifacts from collections after full_history_retention_period.

        :param at: datetime to check if the artifacts are old enough.
        """
        self.get_queryset().exclude(removed_at__isnull=True).exclude(
            artifact__isnull=True
        ).filter(
            # https://github.com/typeddjango/django-stubs/issues/1548
            removed_at__lt=(
                at - F("parent_collection__full_history_retention_period")  # type: ignore[operator] # noqa: E501
            )
        ).update(artifact=None)

    def drop_metadata(self, at: dt.datetime) -> None:
        """
        Delete old collection items.

        After full_history_retention_period + metadata_only_retention_period.

        :param at: datetime to check if the collection item is old enough.
        """
        self.get_queryset().exclude(removed_at__isnull=True).filter(
            # https://github.com/typeddjango/django-stubs/issues/1548
            removed_at__lt=(
                at  # type: ignore[operator]
                - F("parent_collection__full_history_retention_period")
                - F("parent_collection__metadata_only_retention_period")
            )
        ).delete()


class _CollectionItemTypes(models.TextChoices):
    """Choices for the CollectionItem.type."""

    BARE = "b", "Bare"
    ARTIFACT = "a", "Artifact"
    COLLECTION = "c", "Collection"


class CollectionItem(models.Model):
    """CollectionItem model."""

    name = models.CharField(max_length=255)

    Types: TypeAlias = _CollectionItemTypes

    child_type = models.CharField(max_length=1, choices=Types.choices)

    # category duplicates the category of the artifact or collection of this
    # item, so when the underlying artifact or collection is deleted the
    # category is retained
    category = models.CharField(max_length=255)

    parent_collection = models.ForeignKey(
        Collection,
        on_delete=models.PROTECT,
        related_name="child_items",
    )
    parent_category = models.CharField(max_length=255)

    collection = models.ForeignKey(
        Collection,
        on_delete=models.PROTECT,
        related_name="collection_items",
        null=True,
    )
    artifact = models.ForeignKey(
        "Artifact",
        on_delete=models.PROTECT,
        related_name="collection_items",
        null=True,
    )

    data = models.JSONField(default=dict, blank=True)

    created_at = models.DateTimeField(default=timezone.now)
    created_by_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="user_created_%(class)s",
    )
    created_by_workflow = models.ForeignKey(
        "WorkRequest",
        on_delete=models.SET_NULL,
        null=True,
        related_name="workflow_created_%(class)s",
    )

    removed_at = models.DateTimeField(blank=True, null=True)
    removed_by_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        null=True,
        related_name="user_removed_%(class)s",
    )
    removed_by_workflow = models.ForeignKey(
        "WorkRequest",
        on_delete=models.SET_NULL,
        null=True,
        related_name="workflow_removed_%(class)s",
    )

    objects = CollectionItemManager.from_queryset(CollectionItemQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"
        constraints = [
            JsonDataUniqueConstraint(
                fields=[
                    "category",
                    "data->>'codename'",
                    "data->>'architecture'",
                    "data->>'variant'",
                    "data->>'backend'",
                    "parent_collection",
                ],
                condition=Q(
                    parent_category=CollectionCategory.ENVIRONMENTS,
                    removed_at__isnull=True,
                ),
                nulls_distinct=False,
                name="%(app_label)s_%(class)s_unique_debian_environments",
            ),
            UniqueConstraint(
                fields=["name", "parent_collection"],
                condition=Q(removed_at__isnull=True),
                name="%(app_label)s_%(class)s_unique_active_name",
            ),
            # Prevent direct way to add a collection to itself.
            # It is still possible to add loops of collections. The Manager
            # should avoid it
            CheckConstraint(
                check=~Q(collection=F("parent_collection")),
                name="%(app_label)s_%(class)s_distinct_parent_collection",
            ),
            CheckConstraint(
                name="%(app_label)s_%(class)s_childtype_removedat_consistent",
                check=(
                    Q(
                        child_type=_CollectionItemTypes.BARE,
                        collection__isnull=True,
                        artifact__isnull=True,
                    )
                    | (
                        Q(
                            child_type=_CollectionItemTypes.ARTIFACT,
                            collection__isnull=True,
                        )
                        & (
                            Q(artifact__isnull=False)
                            | Q(removed_at__isnull=False)
                        )
                    )
                    | (
                        Q(
                            child_type=_CollectionItemTypes.COLLECTION,
                            artifact__isnull=True,
                        )
                        & (
                            Q(collection__isnull=False)
                            | Q(removed_at__isnull=False)
                        )
                    )
                ),
            ),
        ]
        indexes = [
            models.Index("name", name="%(app_label)s_ci_name_idx"),
            models.Index(
                F("parent_collection"),
                KT("data__package"),
                KT("data__version"),
                name="%(app_label)s_ci_suite_source_idx",
                condition=Q(
                    parent_category=CollectionCategory.SUITE,
                    child_type=_CollectionItemTypes.ARTIFACT,
                    category=ArtifactCategory.SOURCE_PACKAGE,
                ),
            ),
            models.Index(
                F("parent_collection"),
                KT("data__srcpkg_name"),
                KT("data__srcpkg_version"),
                name="%(app_label)s_ci_suite_binary_source_idx",
                condition=Q(
                    parent_category=CollectionCategory.SUITE,
                    child_type=_CollectionItemTypes.ARTIFACT,
                    category=ArtifactCategory.BINARY_PACKAGE,
                ),
            ),
            models.Index(
                F("parent_collection"),
                KT("data__path"),
                name="%(app_label)s_ci_suite_index_path_idx",
                condition=Q(
                    parent_category=CollectionCategory.SUITE,
                    child_type=_CollectionItemTypes.ARTIFACT,
                    category=ArtifactCategory.REPOSITORY_INDEX,
                ),
            ),
            models.Index(
                F("parent_collection"),
                KT("data__path"),
                name="%(app_label)s_ci_archive_index_path_idx",
                condition=Q(
                    parent_category=CollectionCategory.ARCHIVE,
                    child_type=_CollectionItemTypes.ARTIFACT,
                    category=ArtifactCategory.REPOSITORY_INDEX,
                ),
            ),
            # Optimize generating suite indexes.
            models.Index(
                F("parent_collection"),
                KT("data__component"),
                F("created_at"),
                F("removed_at"),
                name="%(app_label)s_ci_suite_all_sources_idx",
                condition=Q(
                    parent_category=CollectionCategory.SUITE,
                    child_type=_CollectionItemTypes.ARTIFACT,
                    category=ArtifactCategory.SOURCE_PACKAGE,
                ),
            ),
            models.Index(
                F("parent_collection"),
                KT("data__component"),
                KT("data__architecture"),
                F("created_at"),
                F("removed_at"),
                name="%(app_label)s_ci_suite_all_binaries_idx",
                condition=Q(
                    parent_category=CollectionCategory.SUITE,
                    child_type=_CollectionItemTypes.ARTIFACT,
                    category=ArtifactCategory.BINARY_PACKAGE,
                ),
            ),
            # Optimize latest:TASKNAME_PACKAGE_ARCHITECTURE lookups in
            # debian:qa-results collections.
            models.Index(
                F("parent_collection"),
                KT("data__task_name"),
                KT("data__package"),
                KT("data__architecture"),
                name="%(app_label)s_ci_qa_results_latest_idx",
                condition=Q(
                    parent_category=CollectionCategory.QA_RESULTS,
                    removed_at__isnull=True,
                ),
            ),
        ]

    @override
    def __str__(self) -> str:
        """Return id, name, collection_id, child_type."""
        item_info = (
            f" Artifact id: {self.artifact.id}"
            if self.artifact
            else (
                f" Collection id: {self.collection.id}"
                if self.collection
                else ""
            )
        )

        return (
            f"Id: {self.id} Name: {self.name} "
            f"Parent collection id: {self.parent_collection_id} "
            f"Child type: {self.child_type}"
            f"{item_info}"
        )

    @override
    def save(self, *args: Any, **kwargs: Any) -> None:
        """Populate `parent_category` on save."""
        if not self.parent_category:
            self.parent_category = self.parent_collection.category
        super().save(*args, **kwargs)

    def get_absolute_url(self) -> str:
        """Return the canonical URL to display the item."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.parent_collection.workspace.scope.name):
            return reverse(
                "workspaces:collections:item_detail",
                kwargs={
                    "wname": self.parent_collection.workspace.name,
                    "ccat": self.parent_collection.category,
                    "cname": self.parent_collection.name,
                    "iid": self.pk,
                    "iname": self.name,
                },
            )

    @staticmethod
    def expand_variables(
        variables: dict[str, str], reference_data: dict[Any, Any]
    ) -> dict[str, str]:
        """
        Expand JSONPath variables against some data.

        The data will normally come from an Artifact.
        """
        jsonpaths = {}
        for name, path in variables.items():
            if name.startswith("$"):
                try:
                    jsonpaths[name[1:]] = jsonpath_rw.parse(path)
                except Exception as e:
                    raise ValueError(e)

        expanded_variables = {}
        for name, jsonpath in jsonpaths.items():
            matches = jsonpath.find(reference_data)
            if len(matches) == 1:
                expanded_variables[name] = matches[0].value
            elif len(matches) > 1:
                raise ValueError(
                    "Too many values expanding",
                    variables[f"${name}"],
                    reference_data,
                )
            else:
                raise KeyError(variables[f"${name}"], reference_data)

        for name, value in variables.items():
            if not name.startswith("$"):
                if name in expanded_variables:
                    raise ValueError(
                        f"Cannot set both '${name}' and '{name}' variables"
                    )
                expanded_variables[name] = value

        return expanded_variables

    @staticmethod
    def expand_name(
        name_template: str, expanded_variables: dict[str, str]
    ) -> str:
        """Format item name following item_template."""
        return name_template.format(**expanded_variables)

    def ui(self, request: "HttpRequest") -> "CollectionItemUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.collections import CollectionItemUI

        return CollectionItemUI.get(request, self)


class CollectionItemMatchConstraint(models.Model):
    """
    Enforce matching-value constraints on collection items.

    All instances of this model with the same :py:attr:`collection`,
    :py:attr:`constraint_name`, and :py:attr:`key` must have the same
    :py:attr:`value`.
    """

    objects: ClassVar[models.Manager["CollectionItemMatchConstraint"]] = (
        models.Manager["CollectionItemMatchConstraint"]()
    )

    collection = models.ForeignKey(
        Collection,
        on_delete=models.CASCADE,
        related_name="item_match_constraints",
    )
    # This is deliberately a bare ID, not a foreign key: some constraints
    # take into account even items that no longer exist but were in a
    # collection in the past.  Also note that the collection item's parent
    # collection may not be the same as this model's collection field: some
    # collections impose indirect constraints via child collections.
    collection_item_id = models.BigIntegerField()
    constraint_name = models.CharField(max_length=255)
    key = models.TextField()
    value = models.TextField()

    class Meta(TypedModelMeta):
        constraints = [
            ExclusionConstraint(
                name="%(app_label)s_%(class)s_match_value",
                expressions=(
                    (F("collection"), RangeOperators.EQUAL),
                    (F("constraint_name"), RangeOperators.EQUAL),
                    (F("key"), RangeOperators.EQUAL),
                    (F("value"), RangeOperators.NOT_EQUAL),
                ),
            )
        ]
        indexes = [
            models.Index(
                name="%(app_label)s_cimc_collection_item_idx",
                fields=["collection_item_id"],
            )
        ]


class _CollectionRelationTypes(models.TextChoices):
    """Types of relations."""

    SUITE_FORKED_FROM = "suite-forked-from", "Forked from"
    SUITE_BASED_ON = "suite-based-on", "Based on"
    SUITE_REQUIRES = "suite-requires", "Requires"
    SUITE_TARGETING = "suite-targeting", "Targeting"
    SUITE_DEFAULT_QA_RESULTS = "suite-default-qa-results", "Default QA results"


class CollectionRelation(models.Model):
    """Model relationships between collections."""

    Types: TypeAlias = _CollectionRelationTypes

    source = models.ForeignKey(
        Collection, on_delete=models.CASCADE, related_name="relations"
    )
    source_category = models.CharField(max_length=255)

    target = models.ForeignKey(
        Collection, on_delete=models.CASCADE, related_name="targeted_by"
    )
    target_category = models.CharField(max_length=255)

    type = models.CharField(
        max_length=32, choices=_CollectionRelationTypes.choices
    )

    position = models.PositiveIntegerField(null=True, blank=True)

    created_at = models.DateTimeField(auto_now_add=True)
    created_by_user = models.ForeignKey(
        "User",
        related_name="created_collection_relations",
        on_delete=models.PROTECT,
    )

    updated_at = models.DateTimeField(null=True, blank=True)
    updated_by_user = models.ForeignKey(
        "User",
        related_name="updated_collection_relations",
        on_delete=models.PROTECT,
        null=True,
    )

    class Meta(TypedModelMeta):
        base_manager_name = "objects"

        constraints = [
            # Enforce that `requires` relations have a position, others not
            models.CheckConstraint(
                name="%(app_label)s_%(class)s_position",
                check=(
                    Q(
                        type=_CollectionRelationTypes.SUITE_REQUIRES,
                        position__isnull=False,
                    )
                    | Q(
                        type__in=[
                            _CollectionRelationTypes.SUITE_FORKED_FROM,
                            _CollectionRelationTypes.SUITE_BASED_ON,
                            _CollectionRelationTypes.SUITE_TARGETING,
                            _CollectionRelationTypes.SUITE_DEFAULT_QA_RESULTS,
                        ],
                        position__isnull=True,
                    )
                ),
            ),
            # Enforce that `requires` relations have unique positions per source
            models.UniqueConstraint(
                name="%(app_label)s_%(class)s_suite_requires_no_duplicate_position",
                fields=["source", "type", "position"],
                # this only applies for SUITE_REQUIRES: the other ones have
                # cardinality max 1 enforced. Not using condition so it can
                # be deferred.
                deferrable=Deferrable.DEFERRED,
            ),
            models.UniqueConstraint(
                name="%(app_label)s_%(class)s_cardinality_checks_max_1",
                fields=["source", "type"],
                condition=(
                    Q(type=_CollectionRelationTypes.SUITE_FORKED_FROM)
                    | Q(type=_CollectionRelationTypes.SUITE_BASED_ON)
                    | Q(type=_CollectionRelationTypes.SUITE_TARGETING)
                    | Q(type=_CollectionRelationTypes.SUITE_DEFAULT_QA_RESULTS)
                ),
            ),
            # Enforce that `requires` relations do not have repeated targets
            models.UniqueConstraint(
                name="%(app_label)s_%(class)s_suite_requires_no_duplicate_target",
                fields=["source", "type", "target"],
                condition=Q(type=_CollectionRelationTypes.SUITE_REQUIRES),
            ),
            models.CheckConstraint(
                name="%(app_label)s_%(class)s_categories_correct",
                check=(
                    # SUITE_FORKED_FROM: both must be suites
                    Q(
                        type=_CollectionRelationTypes.SUITE_FORKED_FROM,
                        source_category=CollectionCategory.SUITE,
                        target_category=CollectionCategory.SUITE,
                    )
                    # SUITE_BASED_ON: both must be suites
                    | Q(
                        type=_CollectionRelationTypes.SUITE_BASED_ON,
                        source_category=CollectionCategory.SUITE,
                        target_category=CollectionCategory.SUITE,
                    )
                    # SUITE_REQUIRES: both must be suites
                    | Q(
                        type=_CollectionRelationTypes.SUITE_REQUIRES,
                        source_category=CollectionCategory.SUITE,
                        target_category=CollectionCategory.SUITE,
                    )
                    # SUITE_TARGETING: both must be suites
                    | Q(
                        type=_CollectionRelationTypes.SUITE_TARGETING,
                        source_category=CollectionCategory.SUITE,
                        target_category=CollectionCategory.SUITE,
                    )
                    # SUITE_DEFAULT_QA_RESULTS: collection must be a suite,
                    # target must be qa-results
                    | Q(
                        type=_CollectionRelationTypes.SUITE_DEFAULT_QA_RESULTS,
                        source_category=CollectionCategory.SUITE,
                        target_category=CollectionCategory.QA_RESULTS,
                    )
                ),
            ),
        ]

    @override
    def __str__(self) -> str:
        """Return string representation of the CollectionRelation."""
        if self.position is not None:
            position_str = f", position {self.position}"
        else:
            position_str = ""

        return (
            f"{self.source.name} → {self.target.name} "
            f"({CollectionRelation.Types(self.type).label}{position_str})"
        )

    @override
    def save(
        self, *args: Any, updated_by_user: "User | None" = None, **kwargs: Any
    ) -> None:
        """Populate denormalized categories and enforce update audit fields."""
        if self.pk is not None:
            # Updating an existing CollectionRelation: check that the User
            # is provided
            if updated_by_user is None:
                raise ValueError(
                    "updated_by_user must be set "
                    "when updating a CollectionRelation"
                )

            if "update_fields" in kwargs:
                # "Model.save.update_fields" is not used in Debusine
                # at the moment. In order to avoid having to deal with
                # "update_fields" not having "update_by_user": just check that
                # is not used in this particular case
                raise ValueError(
                    "update_fields is not supported when updating "
                    "CollectionRelation"
                )

            self.updated_by_user = updated_by_user
            self.updated_at = timezone.now()

        self.position = self._position_for_create(
            source=self.source,
            relation_type=_CollectionRelationTypes(self.type),
            wanted_position=self.position,
        )

        self.source_category = self.source.category
        self.target_category = self.target.category

        with transaction.atomic():
            if self.position is not None:
                CollectionRelation.objects.filter(
                    source=self.source,
                    type=self.type,
                    position__gte=self.position,
                ).update(position=F("position") + 1)

            super().save(*args, **kwargs)

    @permission_check(
        "{user} cannot delete collection relation {resource}",
    )
    def can_delete(self, pc: PermissionContext) -> bool:
        """
        Check if the collection relation can be deleted.

        It can be deleted if the source collection can be configured.
        """
        return self.source.can_configure(pc)

    @classmethod
    def get_max_position(
        cls, collection: Collection, relation_type: _CollectionRelationTypes
    ) -> int:
        """
        Return the current maximum position for a given relation type.

        Return 0 if no relation of this type exists.
        """
        return (
            cls.objects.filter(
                source=collection,
                type=relation_type,
            ).aggregate(max_position=Max("position"))["max_position"]
            or 0
        )

    @classmethod
    def _position_for_create(
        cls,
        *,
        source: Collection,
        relation_type: _CollectionRelationTypes,
        wanted_position: int | None,
    ) -> int | None:
        """
        Return the normalized position for a new relation.

        Keep positions contiguous from 1 to the maximum position.
        """
        if wanted_position is None:
            return None

        current_max_position = cls.get_max_position(
            collection=source, relation_type=relation_type
        )

        if wanted_position < 1:
            return 1
        elif wanted_position > current_max_position + 1:
            return current_max_position + 1
        else:
            return wanted_position

    @override
    def delete(self, *args: Any, **kwargs: Any) -> tuple[int, dict[str, int]]:
        """Delete this relation and shift later positions."""
        with transaction.atomic():
            result = super().delete(*args, **kwargs)

            if self.position is not None:
                CollectionRelation.objects.filter(
                    source=self.source,
                    type=self.type,
                    position__gt=self.position,
                ).update(
                    position=F("position") - 1,
                )

        return result

    def ui(self, request: "HttpRequest") -> "CollectionRelationUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.collection_relations import (
            CollectionRelationUI,
        )

        return CollectionRelationUI.get(request, self)


class CollectionTokenGrantQuerySet[A](QuerySet["CollectionTokenGrant", A]):
    """Custom QuerySet for CollectionTokenGrant."""

    def granted_resources(self, role: RoleBase) -> CollectionQuerySet[Any]:
        """Find collections granted by this queryset with this role."""
        assert isinstance(role, CollectionRoles)
        role_grants = self.filter(role=role)
        return Collection.objects.filter(
            # An explicit grant is enough.
            Q(token_grants__in=role_grants)
            # Granting access to an archive grants access to all its suites.
            | (
                Q(category=CollectionCategory.SUITE)
                & Exists(
                    Collection.objects.filter(
                        category=CollectionCategory.ARCHIVE,
                        workspace=OuterRef("workspace"),
                        token_grants__in=role_grants,
                    )
                )
            )
        )


class CollectionTokenGrantManager(models.Manager["CollectionTokenGrant"]):
    """Manager for the CollectionTokenGrant model."""

    @override
    def get_queryset(self) -> CollectionTokenGrantQuerySet[Any]:
        return CollectionTokenGrantQuerySet(self.model, using=self._db)


class CollectionTokenGrant(models.Model):
    """A grant of a specific role on a collection to a token."""

    collection = models.ForeignKey(
        Collection, on_delete=models.CASCADE, related_name="token_grants"
    )
    token = models.ForeignKey(
        "db.Token", on_delete=models.CASCADE, related_name="collection_grants"
    )
    role = models.CharField(choices=CollectionRoles.choices)

    created_at = models.DateTimeField(default=timezone.now)
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="user_created_%(class)s",
    )

    objects = CollectionTokenGrantManager.from_queryset(
        CollectionTokenGrantQuerySet
    )()

    class Meta(TypedModelMeta):
        constraints = [
            UniqueConstraint(
                fields=["token", "collection", "role"],
                name="%(app_label)s_%(class)s_unique",
            )
        ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Schedule WorkRequests to Workers."""

import logging
from operator import attrgetter
from uuid import uuid4

import django.db
import pgtransaction
from celery import shared_task
from django.conf import settings
from django.db import OperationalError, transaction
from django.db.models import (
    DateTimeField,
    Exists,
    F,
    OuterRef,
    Q,
    QuerySet,
    Window,
)
from django.db.models.fields.json import KT
from django.db.models.functions import Cast, RowNumber
from django.utils import timezone

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts.models import TaskTypes
from debusine.db.context import context
from debusine.db.models import WorkRequest, Worker
from debusine.db.models.work_requests import (
    SkipWorkRequest,
    compute_workflow_last_activity,
    compute_workflow_runtime_status,
    workflow_ancestors,
)
from debusine.db.transactions import atomic
from debusine.project.celery import (
    apply_async_on_commit,
    celery_tasks_in_progress,
)
from debusine.server.celery import run_server_task
from debusine.server.workflows.celery import run_workflow_task
from debusine.tasks import TaskConfigError
from debusine.tasks.models import Confirmation, OutputData, OutputDataError

logger = logging.getLogger(__name__)


class AssignToWorkerError(Exception):
    """Assigning a work request to a worker failed."""

    def __init__(self, message: str, code: str) -> None:
        """Construct the exception."""
        self.message = message
        self.code = code


def _possible_to_schedule_for_worker(worker: Worker) -> bool:
    if worker.is_at_capacity():
        # Worker already has enough work requests assigned; no scheduling
        # needed
        return False

    try:
        Worker.objects.select_for_update(nowait=True).get(id=worker.id)
    except django.db.DatabaseError:
        logger.debug("[SCHEDULER] Failed to lock Worker %s", worker)
        return False

    return True


@transaction.atomic
@context.local()
@context.disable_permission_checks()
def _assign_work_request_to_worker(
    work_request: WorkRequest, worker: Worker, locked_workflow_ids: set[int]
) -> WorkRequest | None:
    """
    Attempt to assign a particular work request to the worker.

    :param work_request: the work request to try to assign.
    :param worker: the worker that needs a new work request to be
      assigned; must already be locked for update.
    :return: The assigned work request, or None if it could not be assigned.
    :raises AssignToWorkerError: if the work request is broken.
    """
    # Enable tag-based scheduling by default for now, but allow to disable it
    # with a feature flag, in case there are issues with it.
    FEATURE_TAG_BASED_SCHEDULING = getattr(
        settings, "FEATURE_TAG_BASED_SCHEDULING", True
    )

    try:
        task = work_request.get_task()
    except TaskConfigError as exc:
        raise AssignToWorkerError(
            f"Failed to configure: {exc}", "configure-failed"
        ) from exc

    if not FEATURE_TAG_BASED_SCHEDULING and not task.can_run_on(
        worker.metadata()
    ):
        return None

    try:
        # We don't need to try to lock the whole workflow here, since
        # WorkRequest.assign_worker has no cascading effects on other work
        # requests.
        work_request = WorkRequest.objects.select_for_update(nowait=True).get(
            id=work_request.id
        )
    except django.db.DatabaseError:
        if work_request.workflow_root_id is not None:
            # The whole workflow is probably locked, so don't spam logs
            # trying each pending work request in it.
            locked_workflow_ids.add(work_request.workflow_root_id)
        logger.debug(
            "[SCHEDULER] Failed to lock WorkRequest %s"
            " (skipping this workflow)",
            work_request,
        )
        transaction.set_rollback(True)
        return None

    if work_request.worker:  # pragma: no cover
        # work_request did not have a worker assigned on the
        # initial qs but has one now - nothing to do
        return None

    try:
        work_request.assign_worker(worker)
    except SkipWorkRequest:
        logger.debug(
            "Skipping work request %s due to on_assignment action", work_request
        )
        return None
    except Exception as exc:
        raise AssignToWorkerError(
            f"Failed to compute dynamic data: {exc}", "dynamic-data-failed"
        ) from exc

    if task.TASK_TYPE is TaskTypes.SERVER:
        # Use a slightly more informative task ID than the default.
        task_id = f"{work_request.task_name}_{work_request.id}_{uuid4()}"
        apply_async_on_commit(
            run_server_task, args=(work_request.id,), task_id=task_id
        )

    return work_request


def _log_work_request_assign_error(
    work_request: WorkRequest, exc: AssignToWorkerError
) -> None:
    """
    Log an error arisen when assigning a work request to a worker.

    This is in a separate function to make it easier to mock.
    """
    logger.error(
        "Error assigning work request %s/%s (%s): %s",
        work_request.task_type,
        work_request.task_name,
        work_request.id,
        exc.message,
    )


def _assign_work_request_to_worker_or_error(
    work_request: WorkRequest, worker: Worker, locked_workflow_ids: set[int]
) -> WorkRequest | None:
    """
    Attempt to assign a particular work request to the worker, handling errors.

    This needs special care around transaction management.  Assigning the
    worker shouldn't take a whole-workflow lock, because it doesn't need to
    and doing so would increase contention; but if the work request is
    broken then we must take a whole-workflow lock because marking the work
    request as completed might cascade to other work requests in the same
    workflow, and that must therefore be done in a separate savepoint from
    the assignment to avoid a potential deadlock.

    :param work_request: the work request to try to assign.
    :param worker: the worker that needs a new work request to be
      assigned; must already be locked for update.
    :return: The assigned work request, or None if it could not be assigned.
    """
    try:
        return _assign_work_request_to_worker(
            work_request, worker, locked_workflow_ids
        )
    except AssignToWorkerError as exc:
        with transaction.atomic():
            # _assign_work_request_to_worker runs in its own savepoint, so
            # any update locks it holds will be released in case an
            # exception is raised, allowing this to be called in the
            # exception handler without risking a deadlock on WorkRequest
            # locks.  We must still use nowait=True, both because we're in
            # the scheduler and can't block, and because we're within a
            # Worker update lock and don't want to risk something else
            # trying to acquire WorkRequest and Worker locks in the opposite
            # order.
            try:
                WorkRequest.objects.filter(
                    id=work_request.id
                ).lock_workflows_for_update(nowait=True)
            except django.db.DatabaseError:
                # Not ideal, but we'll try the same work request again
                # later.
                logger.debug(
                    "[SCHEDULER] Failed to lock WorkRequest %s to record error",
                    work_request,
                )
                transaction.set_rollback(True)
            else:
                _log_work_request_assign_error(work_request, exc)
                work_request.mark_completed(
                    WorkRequest.Results.ERROR,
                    output_data=OutputData(
                        errors=[
                            OutputDataError(message=exc.message, code=exc.code)
                        ]
                    ),
                )
        return None


def get_provided_worker_tags(worker: Worker) -> set[str]:
    """Get the set of tags provided by a worker."""
    from debusine.server.tags import make_worker_provided_tagset

    tagset = make_worker_provided_tagset()
    tagset.add(
        wtags.ProvenanceProvided.WORKER,
        worker.dynamic_metadata.get("provided_tags"),
    )
    tagset.add(
        wtags.ProvenanceProvided.ADMIN,
        worker.static_metadata.get("provided_tags"),
    )
    tagset.add(
        wtags.ProvenanceProvided.SYSTEM,
        {wtags.WORKER_TYPE_PREFIX + worker.worker_type},
    )
    tagset.finalize()
    return tagset.tags


def get_required_task_tags(worker: Worker) -> set[str]:
    """Get the set of task tags required by a worker."""
    from debusine.server.tags import make_worker_required_tagset

    tagset = make_worker_required_tagset()
    tagset.add(
        ttags.ProvenanceRequired.ADMIN,
        worker.static_metadata.get("required_tags"),
    )
    tagset.finalize()
    return tagset.tags


def get_candidate_work_requests(worker: Worker) -> QuerySet[WorkRequest]:
    """Get pending work requests that can be scheduled on this worker."""
    # Disable tag-based scheduling by default for now, but allow to enable it
    # with a feature flag so we can test it, before offering it for mainstream
    # adoption
    FEATURE_TAG_BASED_SCHEDULING = getattr(
        settings, "FEATURE_TAG_BASED_SCHEDULING", True
    )

    provided_worker_tags = get_provided_worker_tags(worker)
    required_task_tags = get_required_task_tags(worker)
    worker_metadata = worker.metadata()
    tasks_allowlist = worker_metadata.get("tasks_allowlist", None)
    tasks_denylist = worker_metadata.get("tasks_denylist", [])

    # TODO: Optimization: pre-filter on task type as appropriate.
    qs = WorkRequest.objects.pending(exclude_assigned=True).filter(
        task_type__in=(
            TaskTypes.WORKER,
            TaskTypes.SERVER,
            TaskTypes.SIGNING,
        )
    )
    if tasks_allowlist is not None:
        qs = qs.filter(task_name__in=tasks_allowlist)
    elif tasks_denylist:
        qs = qs.exclude(task_name__in=tasks_denylist)

    if FEATURE_TAG_BASED_SCHEDULING:
        qs = qs.filter(
            scheduler_tags_required__contained_by=list(provided_worker_tags),
            scheduler_tags_provided__contains=list(required_task_tags),
        )

    return qs


@transaction.atomic
def schedule_for_worker(
    worker: Worker, locked_workflow_ids: set[int]
) -> WorkRequest | None:
    """
    Schedule a new work request for the worker.

    :param worker: the worker that needs a new work request to be
      assigned.
    :param locked_workflow_ids: a set of workflows that have so far
      discovered to have been locked.
    :return: The assigned work request. None if no suitable work request could
      be found.
    :rtype: WorkRequest | None.
    """
    if not _possible_to_schedule_for_worker(worker):
        return None

    for work_request in get_candidate_work_requests(worker):
        if work_request.workflow_root_id in locked_workflow_ids:
            continue
        assigned_work_request = _assign_work_request_to_worker_or_error(
            work_request, worker, locked_workflow_ids
        )
        if assigned_work_request is not None:
            return assigned_work_request

    return None


# We may encounter serialization failures, but if so it's OK to just let the
# next attempt catch up.
@atomic(isolation_level=pgtransaction.REPEATABLE_READ)
def update_workflows() -> None:
    """
    Update expensive properties of workflows.

    The ``workflow_last_activity_at`` and ``workflow_runtime_status`` fields
    of workflows are used in the web UI and need to be kept up to date.
    """
    try:
        need_update = WorkRequest.objects.filter(
            workflows_need_update=True
        ).lock_workflows_for_update()
    except OperationalError as exc:
        # Something like select_for_update(skip_locked=True) might be
        # better, allowing us to make some progress, but it's unclear how to
        # acquire a lock on only those work requests that are in workflows
        # that can be locked as a whole without blocking.
        logger.warning(
            "update_workflows() deferred due to a concurrent update: %s",
            exc,
        )
        return
    for workflow in workflow_ancestors(need_update):
        assert workflow.task_type == TaskTypes.WORKFLOW
        workflow.workflow_last_activity_at = compute_workflow_last_activity(
            workflow
        )
        workflow.workflow_runtime_status = compute_workflow_runtime_status(
            workflow
        )
        workflow.save()
    need_update.update(workflows_need_update=False)


def run_workflow_task_via_celery(work_request: WorkRequest) -> None:
    """Run a workflow callback or workflow via Celery."""
    # Use a slightly more informative task ID than the default.
    task_id = (
        f"{work_request.task_type.lower()}_{work_request.task_name}_"
        f"{work_request.id}_{uuid4()}"
    )
    apply_async_on_commit(
        run_workflow_task, args=(work_request.id,), task_id=task_id
    )


@transaction.atomic
def schedule_internal() -> list[WorkRequest]:
    """
    Schedule internal work requests.

    These don't require a worker, but need some kind of action from the
    scheduler.
    """
    in_progress_task_results = celery_tasks_in_progress(
        "debusine.server.workflows.celery.run_workflow_task"
    )

    pending_wrs = WorkRequest.objects.pending()
    pending_synchronization_points = pending_wrs.synchronization_points()
    pending_workflow_callbacks = (
        pending_wrs.workflow_callbacks()
        .annotate(
            row_number=Window(
                RowNumber(), partition_by=F("parent"), order_by="id"
            )
        )
        # Only one workflow callback for a given workflow may run at once.
        .exclude(
            Exists(
                in_progress_task_results.filter(
                    first_task_arg__in=WorkRequest.objects.filter(
                        parent=OuterRef(OuterRef("parent"))
                    ),
                )
            )
        )
        .exclude(row_number__gt=1)
    )
    pending_workflows = (
        pending_wrs.filter(task_type=TaskTypes.WORKFLOW)
        .annotate(
            row_number=Window(
                RowNumber(), partition_by=F("workflow_root"), order_by="id"
            )
        )
        # Only one sub-workflow within a given root workflow may run at once.
        .exclude(
            Exists(
                in_progress_task_results.filter(
                    first_task_arg=OuterRef("workflow_root")
                )
            )
        )
        .exclude(row_number__gt=1)
    )
    pending_waits = pending_wrs.filter(task_type=TaskTypes.WAIT)
    pending_combined = WorkRequest.objects.filter(
        Q(id__in=pending_synchronization_points)
        | Q(id__in=pending_workflow_callbacks)
        | Q(id__in=pending_workflows)
        | Q(id__in=pending_waits)
    )

    try:
        pending_combined.lock_workflows_for_update(nowait=True)
    except django.db.DatabaseError:
        # Something like select_for_update(skip_locked=True) might be
        # better, allowing us to make some progress, but it's unclear how to
        # acquire a lock on only those work requests that are in workflows
        # that can be locked as a whole without blocking.  This should be
        # fine provided that nothing holds a lock for too long on workflows
        # that had pending internal work requests at the start of its
        # transaction.
        logger.debug(
            "[SCHEDULER] Failed to lock workflows with pending internal"
            " work requests"
        )
        return []

    workflow_roots: set[WorkRequest] = set()
    result: list[WorkRequest] = []
    for work_request in pending_combined:
        match (work_request.task_type, work_request.task_name):
            case (TaskTypes.INTERNAL, "synchronization_point"):
                # Synchronization points do nothing; when pending, they are
                # immediately marked as completed, thus unblocking work
                # requests that depend on them.
                work_request.mark_completed(WorkRequest.Results.SUCCESS)
                result.append(work_request)

            case (TaskTypes.INTERNAL, "workflow"):
                # Workflow callbacks are handled by the corresponding
                # workflow orchestrator.  This may take a while, so we
                # delegate it to a Celery task.
                work_request.mark_running()
                run_workflow_task_via_celery(work_request)
                result.append(work_request)

            case (TaskTypes.WORKFLOW, _):
                # Workflow population always happens via the orchestrator
                # for the root workflow.  Sub-workflow orchestrators are not
                # called directly from the scheduler, although if a
                # sub-workflow is pending then its root workflow may deal
                # with populating it.
                # Workflows always have a root.
                assert work_request.workflow_root is not None
                workflow_roots.add(work_request.workflow_root)

            case (TaskTypes.WAIT, _):
                # Wait tasks are marked running immediately, but otherwise
                # we do nothing; some later event will mark them as
                # completed.
                work_request.mark_running()
                result.append(work_request)

            case _ as unreachable:
                raise AssertionError(
                    f"Unexpected internal task name: {unreachable}"
                )

    for workflow_root in sorted(workflow_roots, key=attrgetter("id")):
        # Workflows are populated by the corresponding workflow
        # orchestrator.  This may take a while, so we delegate it to a
        # Celery task.
        workflow_root.mark_running()
        run_workflow_task_via_celery(workflow_root)
        result.append(workflow_root)

    return result


@transaction.atomic
def complete_wait_tasks() -> None:
    """Mark certain wait tasks as completed."""
    qs = (
        WorkRequest.objects.running()
        .filter(task_type=TaskTypes.WAIT)
        .filter(
            # Delay tasks complete automatically once their delay has
            # expired.
            Q(task_name="delay")
            # Confirm tasks with auto_confirm_if_no_failure=True complete
            # automatically if all their dependencies are successful.
            | (
                Q(
                    task_name="confirm",
                    task_data__auto_confirm_if_no_failure=True,
                )
                & ~Exists(
                    WorkRequest.objects.filter(
                        Q(reverse_dependencies=OuterRef("id"))
                        & ~Q(
                            status=WorkRequest.Statuses.COMPLETED,
                            result=WorkRequest.Results.SUCCESS,
                        )
                    )
                )
            )
        )
        .annotate(
            delay_until=Cast(
                KT("task_data__delay_until"),
                DateTimeField(blank=True, null=True),
            )
        )
        .exclude(task_name="delay", delay_until__gt=timezone.now())
    )

    try:
        qs.lock_workflows_for_update(nowait=True)
    except django.db.DatabaseError:
        # Something like select_for_update(skip_locked=True) might be
        # better, allowing us to make some progress, but it's unclear how to
        # acquire a lock on only those work requests that are in workflows
        # that can be locked as a whole without blocking.  This should be
        # fine provided that nothing holds a lock for too long on workflows
        # that had pending internal work requests at the start of its
        # transaction.
        logger.debug(
            "[SCHEDULER] Failed to lock workflows with pending internal"
            " work requests"
        )
        return

    for work_request in qs:
        match work_request.task_name:
            case "delay":
                work_request.mark_completed(WorkRequest.Results.SUCCESS)
            case "confirm":
                work_request.mark_completed(
                    WorkRequest.Results.SUCCESS,
                    output_data=OutputData(
                        confirmation=Confirmation(confirmed=True)
                    ),
                )
            case _ as unreachable:
                raise AssertionError(f"Unexpected task name: {unreachable}")


def schedule() -> list[WorkRequest]:
    """
    Perform work request scheduling.

    This runs frequently, and does a few different steps:

    * Keep various properties of workflows up to date.
    * Check pending work requests that can be acted on directly without a
      worker, as well as pending workflow instances and workflow callbacks
      that need to be orchestrated by a Celery task.
    * Loop over idle workers and try to assign pending work requests to
      them.

    :return: the list of work requests that got assigned.
    """
    update_workflows()

    available_workers = Worker.objects.waiting_for_work_request()

    result = list(schedule_internal())
    locked_workflow_ids: set[int] = set()
    for available_worker in available_workers:
        work_request = schedule_for_worker(
            available_worker, locked_workflow_ids
        )
        if work_request:
            result.append(work_request)

    complete_wait_tasks()

    return result


# mypy complains that celery.shared_task is untyped, which is true, but we
# can't fix that here.
@shared_task()  # type: ignore[untyped-decorator]
def schedule_task() -> None:
    """Run the scheduler as a celery task."""
    schedule()


def schedule_soon() -> None:
    """Ask the scheduler to run soon."""
    if not getattr(settings, "DISABLE_AUTOMATIC_SCHEDULING", False):
        apply_async_on_commit(schedule_task, once_per_savepoint=True)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the views."""

from typing import Any
from unittest import mock

from django.conf import settings
from django.contrib import messages
from django.contrib.messages.storage.base import Message
from django.urls import reverse
from rest_framework import status

from debusine.db.models import WorkRequest, Workspace
from debusine.db.playground import scenarios
from debusine.test.django import ListFilter, TestCase, override_permission
from debusine.web.views import HomepageView
from debusine.web.views.tests.utils import ViewTestMixin, html_check_icon


class HomepageViewTests(ViewTestMixin, TestCase):
    """Tests for the Homepage class."""

    scenario = scenarios.DefaultScopeUser()

    xpath_scope_workspace_list = "//table[@id='scope-workspace-list']"
    xpath_current_workflow_list = "//div[@id='current-workflow-list-table']"
    xpath_completed_workflow_list = "//div[@id='completed-workflow-list-table']"

    def test_html_check_icon(self) -> None:
        """Test that html_check_icon returns the right results."""
        self.assertIn("green", html_check_icon(True))
        self.assertNotIn("green", html_check_icon(False))
        self.assertIn("red", html_check_icon(False))
        self.assertNotIn("red", html_check_icon(True))

    def test_slash(self) -> None:
        """
        Slash (/) URL and homepage are the same.

        Make sure that the / is handled by the 'homepage' view.
        """
        self.assertEqual(reverse("homepage:homepage"), "/")

    def test_homepage(self) -> None:
        """Homepage view loads (not logged in)."""
        list_your_tokens_url = reverse(
            "user:token-list", kwargs={"username": self.scenario.user.username}
        )
        list_your_tokens_html = (
            f'<a class="dropdown-item" href="{list_your_tokens_url}">Tokens</a>'
        )

        response = self.client.get(reverse("homepage:homepage"))
        tree = self.assertResponseHTML(response)
        h1 = self.assertHasElement(tree, "//h1")
        self.assertTextContentEqual(h1, "Welcome to Debusine!")

        title = self.assertHasElement(tree, "//title")
        self.assertTextContentEqual(title, "Main page")

        self.assertNavCommonElements(tree)
        self.assertNavNoUser(tree)

        self.assertHasElement(tree, self.xpath_scope_workspace_list)

        self.assertNotContains(response, list_your_tokens_html, html=True)
        self.assertFalse(tree.xpath(self.xpath_current_workflow_list))
        self.assertFalse(tree.xpath(self.xpath_completed_workflow_list))

    def test_homepage_logged_in(self) -> None:
        """User is logged in: contains "You are authenticated as: username"."""
        self.client.force_login(self.scenario.user)
        response = self.client.get(reverse("homepage:homepage"))
        tree = self.assertResponseHTML(response)

        self.assertNavCommonElements(tree)
        self.assertNavHasUser(tree, self.scenario.user)

        table = self.assertHasElement(tree, self.xpath_scope_workspace_list)
        tbody = self.assertHasElement(table, "tbody")
        self.assertEqual(len(tbody.tr), 1)
        tr = tbody.tr[0]
        self.assertTextContentEqual(tr.td[0], self.scenario.scope.name)
        self.assertTextContentEqual(tr.td[1], "System")
        self.assertTextContentEqual(tr.td[2], "Viewer")
        self.assertTextContentEqual(tr.td[3], "0")
        self.assertTextContentEqual(tr.td[4], "0")
        self.assertTextContentEqual(tr.td[5], "0")

        table = self.assertHasElement(tree, self.xpath_current_workflow_list)
        table = self.assertHasElement(tree, self.xpath_completed_workflow_list)

    def test_messages(self) -> None:
        """Messages from django.contrib.messages are displayed."""

        def mocked_get_context_data(
            self: HomepageView, **kwargs: Any
        ) -> dict[str, Any]:
            messages.error(self.request, "Error message")
            return {
                "base_template": HomepageView.base_template,
                "workspaces": "",
            }

        with mock.patch(
            "debusine.web.views.views.HomepageView.get_context_data",
            autospec=True,
            side_effect=mocked_get_context_data,
        ):
            response = self.client.get(reverse("homepage:homepage"))

        tree = self.assertResponseHTML(response)
        div = self.assertHasElement(tree, "//div[@id='user-message-container']")
        msgdiv = div.div.div
        self.assertEqual(msgdiv.get("class"), "toast")
        self.assertEqual(msgdiv.get("role"), "alert")
        assert msgdiv.div[1].text is not None
        self.assertEqual(msgdiv.div[1].text.strip(), "Error message")

    def test_homepage_workflow_list(self) -> None:
        """Homepage lists visible scopes."""
        self.client.force_login(self.scenario.user)
        scope1 = self.playground.get_or_create_scope("scope1")
        workspace1 = self.playground.create_workspace(
            name="workspace1", scope=scope1
        )
        scope2 = self.playground.get_or_create_scope("scope2")
        self.playground.create_workspace(name="workspace2", scope=scope2)

        with override_permission(
            Workspace, "can_display", ListFilter, exclude=[workspace1]
        ):
            response = self.client.get(reverse("homepage:homepage"))
        tree = self.assertResponseHTML(response)
        table = self.assertHasElement(tree, self.xpath_scope_workspace_list)
        tbody = self.assertHasElement(table, "tbody")
        actual = [
            (
                self.get_node_text_normalized(tr.td[0]),
                (self.get_node_text_normalized(tr.td[1])),
            )
            for tr in tbody.tr
        ]
        self.assertEqual(
            actual,
            [
                ("debusine", "System"),
                ("scope2", "workspace2"),
            ],
        )

    def test_queries_anonymous(self) -> None:
        """Test the number of queries for an anonymous visit."""
        with self.assertNumQueries(5):
            self.client.get(reverse("homepage:homepage"))

    def test_queries_anonymous_with_workspaces(self) -> None:
        """Test the number of queries for an anonymous visit."""
        for i in range(3):
            self.playground.create_workspace(name=f"workspace{i}")
        with self.assertNumQueries(5):
            self.client.get(reverse("homepage:homepage"))

    def test_queries_authenticated(self) -> None:
        """Test the number of queries for an authenticated visit."""
        self.client.force_login(self.scenario.user)
        with self.assertNumQueries(11):
            self.client.get(reverse("homepage:homepage"))

    def test_queries_authenticated_with_workspaces(self) -> None:
        """Test the number of queries for an anonymous visit."""
        self.client.force_login(self.scenario.user)
        for i in range(3):
            self.playground.create_workspace(name=f"workspace{i}")
        with self.assertNumQueries(11):
            self.client.get(reverse("homepage:homepage"))

    def test_queries_authenticated_with_workflows(self) -> None:
        """Test query count for a visit with work requests shown."""
        self.client.force_login(self.scenario.user)
        for i in range(3):
            self.playground.create_workspace(name=f"workspace{i}")
        template = self.playground.create_workflow_template("test", "noop")
        for _ in range(5):
            workflow = self.playground.create_workflow(template)
            workflow.mark_running()
        with self.assertNumQueries(14):
            response = self.client.get(reverse("homepage:homepage"))
        self.assertEqual(
            len(response.context["workflows_current"].page_obj.object_list),
            5,
        )

    def test_homepage_customization(self) -> None:
        """Homepage extended by templates installed by site admins."""
        response = self.client.get(reverse("homepage:homepage"))
        tree = self.assertResponseHTML(response)
        h1 = self.assertHasElement(tree, "//h1")
        self.assertTextContentEqual(h1, "Welcome to Debusine!")

        with self.custom_template("web/homepage.html") as template:
            template.write_text(
                "{% extends 'web/homepage.html' %}"
                "{% block pagetitle %}"
                "<h1 class='jumbo'>Welcome to a custom Debusine</h1>"
                "{% endblock %}"
            )
            response = self.client.get(reverse("homepage:homepage"))
            tree = self.assertResponseHTML(response)
            h1 = self.assertHasElement(tree, "//h1[@class='jumbo']")
            self.assertTextContentEqual(h1, "Welcome to a custom Debusine")

    def test_only_root_workflows_shown(self) -> None:
        self.client.force_login(self.scenario.user)
        template = self.playground.create_workflow_template("test", "noop")
        root_workflow = self.playground.create_workflow(template)
        root_workflow.mark_running()
        child_workflow = self.playground.create_workflow(
            template, parent=root_workflow
        )
        child_workflow.mark_running()
        completed_child_workflow = self.playground.create_workflow(
            template, parent=root_workflow
        )
        completed_child_workflow.mark_running()
        completed_child_workflow.mark_completed(
            result=WorkRequest.Results.SUCCESS
        )
        response = self.client.get(reverse("homepage:homepage"))
        self.assertQuerySetEqual(
            response.context["workflows_current"].page_obj.object_list,
            [root_workflow],
        )

        workspaces_qs = response.context["workspaces"].table.rows
        self.assertEqual(workspaces_qs.count(), 1)
        workspace = workspaces_qs.first()
        self.assertEqual(
            (workspace.running, workspace.needs_input, workspace.completed),
            (1, None, None),
        )

    def test_workflows_completed_includes_aborted(self) -> None:
        template = self.playground.create_workflow_template("test", "noop")
        running_workflow = self.playground.create_workflow(template)
        running_workflow.mark_running()
        completed_workflow = self.playground.create_workflow(template)
        completed_workflow.mark_running()
        completed_workflow.mark_completed(result=WorkRequest.Results.SUCCESS)
        aborted_workflow = self.playground.create_workflow(template)
        aborted_workflow.mark_running()
        aborted_workflow.mark_aborted()

        self.client.force_login(self.scenario.user)
        response = self.client.get(reverse("homepage:homepage"))
        self.assertQuerySetEqual(
            response.context["workflows_completed"].page_obj.object_list,
            [completed_workflow, aborted_workflow],
            ordered=False,
        )

        workspaces_qs = response.context["workspaces"].table.rows
        self.assertEqual(workspaces_qs.count(), 1)
        workspace = workspaces_qs.first()
        self.assertEqual(
            (workspace.running, workspace.needs_input, workspace.completed),
            (1, None, 2),
        )


class AdminTests(TestCase):
    """Test the admin view."""

    def assertLoginRequired(self, url: str) -> None:
        """Test that a login is required to access the given URL."""
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        # TODO: we apparently have an admin:login view: what does it do? Does
        # it need to be secured? Do we want an admin at all?
        self.assertEqual(
            response["Location"],
            reverse("admin:login") + "?next=" + url,
        )

    def test_admin_url(self) -> None:
        """Test resolving the admin URL."""
        url = reverse("admin:index")
        self.assertTrue(url.startswith("/-/admin"))

    def test_admin_anonymous_user(self) -> None:
        """Test resolving the admin URL."""
        self.assertLoginRequired(reverse("admin:index"))

    def test_admin_ordinary_user(self) -> None:
        """Test resolving the admin URL."""
        self.client.force_login(self.playground.get_default_user())
        self.assertLoginRequired(reverse("admin:index"))

    def test_admin_staff(self) -> None:
        """Test resolving the admin URL."""
        user = self.playground.get_default_user()
        user.is_staff = True
        user.save()
        self.client.force_login(user)
        response = self.client.get(reverse("admin:index"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_admin_superuser(self) -> None:
        """Test resolving the admin URL."""
        user = self.playground.get_default_user()
        user.is_superuser = True
        user.save()
        self.client.force_login(user)
        self.assertLoginRequired(reverse("admin:index"))


class LegacyRedirectTests(TestCase):
    """Test the best effort redirects for unscoped URLs."""

    def test_scope_redirect_valid(self) -> None:
        """Test redirection of a valid legacy URL."""
        response = self.client.get("/accounts/foo/bar?baz=true")
        self.assertEqual(
            response.status_code, status.HTTP_301_MOVED_PERMANENTLY
        )
        self.assertEqual(
            response.headers["Location"],
            f"/{settings.DEBUSINE_DEFAULT_SCOPE}/accounts/foo/bar?baz=true",
        )

    def test_scope_redirect_invalid(self) -> None:
        """Test redirection of an invalid legacy URL."""
        # Misspelled url (misses trailing 's')
        response = self.client.get("/account/foo/bar?baz=true")
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_scope_reproduce_redirect_loop(self) -> None:
        """Reproduce a redirect loop."""
        artifact, _ = self.playground.create_artifact()

        response = self.client.get(f"/api/1.0/artifact/{artifact.pk}")
        self.assertEqual(
            response.status_code, status.HTTP_301_MOVED_PERMANENTLY
        )
        self.assertEqual(
            response.headers["Location"], f"/api/1.0/artifact/{artifact.pk}/"
        )

        response = self.client.get(
            f"/{settings.DEBUSINE_DEFAULT_SCOPE}/System/artifact/{artifact.pk}"
        )
        self.assertEqual(
            response.status_code, status.HTTP_301_MOVED_PERMANENTLY
        )
        self.assertEqual(
            response.headers["Location"],
            f"/{settings.DEBUSINE_DEFAULT_SCOPE}"
            f"/System/artifact/{artifact.pk}/",
        )

    # urlconfs are created at Django startup, so changing the setting
    # afterwards has no effect: the redirect replacement pattern has already
    # been created. This test would be pointless, but keeping it here as
    # documentation. I would not subclass RedirectView only to have a more
    # dynamic replacement pattern for use only in tests.
    #
    # @override_settings(DEBUSINE_DEFAULT_SCOPE="debian")
    # def test_redirect_different_scope(self) -> None:
    #     """Redirection uses DEBUSINE_DEFAULT_SCOPE."""
    #     response = self.client.get("/accounts/foo/bar?baz=true")
    #     self.assertEqual(
    #         response.status_code, status.HTTP_301_MOVED_PERMANENTLY
    #     )
    #     self.assertEqual(
    #         response.headers["Location"],
    #         "/debian/accounts/foo/bar?baz=true",
    #     )

    def test_workspace_redirect_valid(self) -> None:
        """Test redirection of a valid legacy workspace URL."""
        response = self.client.get("/debusine/workspace/foo/bar?baz=true")
        self.assertEqual(
            response.status_code, status.HTTP_301_MOVED_PERMANENTLY
        )
        self.assertEqual(
            response.headers["Location"],
            "/debusine/foo/bar?baz=true",
        )

    def test_redirect_task_status(self) -> None:
        """Test redirection on the /task-status/ URL."""
        response = self.client.get("/task-status/foo/bar?baz=true")
        self.assertRedirects(
            response,
            "/-/status/queue/foo/bar?baz=true",
            status_code=status.HTTP_301_MOVED_PERMANENTLY,
            fetch_redirect_response=False,
        )

    def test_redirect_workers(self) -> None:
        """Test redirection on the /workers/ URL."""
        response = self.client.get("/workers/foo/bar?baz=true")
        self.assertRedirects(
            response,
            "/-/status/workers/foo/bar?baz=true",
            status_code=status.HTTP_301_MOVED_PERMANENTLY,
            fetch_redirect_response=False,
        )

    def test_redirect_user(self) -> None:
        """Test redirection on the /user/ URL."""
        response = self.client.get("/user/foo/bar?baz=true")
        self.assertRedirects(
            response,
            "/-/user/foo/bar?baz=true",
            status_code=status.HTTP_301_MOVED_PERMANENTLY,
            fetch_redirect_response=False,
        )

    def test_redirect_signon(self) -> None:
        """Test redirection on the signon URLs."""
        scope = settings.DEBUSINE_DEFAULT_SCOPE
        for path, view_name in (
            ("/accounts/oidc_callback/{name}/", "signon:oidc_callback"),
            ("/accounts/bind_identity/{name}/", "signon:bind_identity"),
            ("/{scope}/accounts/oidc_callback/{name}/", "signon:oidc_callback"),
            ("/{scope}/accounts/bind_identity/{name}/", "signon:bind_identity"),
        ):
            for name in ("foo", "bar"):
                url = path.format(name=name, scope=scope)
                with self.subTest(url=url):
                    response = self.client.get(url)
                    self.assertRedirects(
                        response,
                        reverse(view_name, kwargs={"name": name}),
                        status_code=status.HTTP_301_MOVED_PERMANENTLY,
                        fetch_redirect_response=False,
                    )


class ToastViewTests(ViewTestMixin, TestCase):
    """Tests for the :py:class:`ToastView` class."""

    title = "Test showing django messages"

    def get(self, type_arg: str | None = None) -> list[Message]:
        data: dict[str, Any] = {}
        if type_arg is not None:
            data["type"] = type_arg
        response = self.client.get(reverse("test-toasts"), data)
        self.assertResponseHTML(response)
        return list(messages.get_messages(response.wsgi_request))

    def test_types(self) -> None:
        for type_arg, expected_levels in (
            (None, []),
            ("debug", []),
            ("info", ["info"]),
            ("success", ["success"]),
            ("warning", ["warning"]),
            ("error", ["error"]),
            ("invalid", ["error"]),
        ):
            with self.subTest(type_arg=type_arg):
                msgs = self.get(type_arg)
                self.assertEqual([m.level_tag for m in msgs], expected_levels)


class ThemeSelectionViewTests(ViewTestMixin, TestCase):
    """Tests for the :py:class:`ThemeSelectionView` class."""

    def test_get(self) -> None:
        response = self.client.get(reverse("theme-selection"))
        self.assertEqual(
            response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED
        )

    def test_set(self) -> None:
        for value in ("light", "dark", "foo"):
            with self.subTest(value=value):
                response = self.client.post(
                    reverse("theme-selection"), data={"theme": value}
                )
                self.assertRedirects(response, reverse("homepage:homepage"))
                self.assertEqual(self.client.cookies["theme"].value, value)

    def test_set_auto(self) -> None:
        response = self.client.post(
            reverse("theme-selection"), data={"theme": "foo"}
        )
        self.assertRedirects(response, reverse("homepage:homepage"))
        self.assertEqual(self.client.cookies["theme"].value, "foo")
        response = self.client.post(
            reverse("theme-selection"), data={"theme": "auto"}
        )
        self.assertRedirects(response, reverse("homepage:homepage"))
        self.assertEqual(self.client.cookies["theme"].value, "")

    def test_set_redirect(self) -> None:
        response = self.client.post(
            reverse("theme-selection"),
            data={"theme": "foo", "next": "http://example.org"},
        )
        self.assertRedirects(
            response, "http://example.org", fetch_redirect_response=False
        )
        self.assertEqual(self.client.cookies["theme"].value, "foo")
